@@ -373,10 +373,6 @@ static void hfp_dial(struct ofono_voicecall *vc,
 	else
 		snprintf(buf, sizeof(buf), "ATD%s", ph->number);
 
-	if ((clir != OFONO_CLIR_OPTION_DEFAULT) ||
-			(cug != OFONO_CUG_OPTION_DEFAULT))
-		goto error;
-
 	strcat(buf, ";");
 
 	if (g_at_chat_send(vd->chat, buf, none_prefix,
@@ -384,8 +380,7 @@ static void hfp_dial(struct ofono_voicecall *vc,
 		return;
 
 error:
-	if (cbd)
-		g_free(cbd);
+	g_free(cbd);
 
 	CALLBACK_WITH_FAILURE(cb, data);
 }
@@ -410,8 +405,7 @@ static void hfp_template(const char *cmd, struct ofono_voicecall *vc,
 		return;
 
 error:
-	if (req)
-		g_free(req);
+	g_free(req);
 
 	CALLBACK_WITH_FAILURE(cb, data);
 }
@@ -527,8 +521,7 @@ static void hfp_release_specific(struct ofono_voicecall *vc, int id,
 		return;
 
 error:
-	if (req)
-		g_free(req);
+	g_free(req);
 
 	CALLBACK_WITH_FAILURE(cb, data);
 }
@@ -616,8 +609,7 @@ static void hfp_send_dtmf(struct ofono_voicecall *vc, const char *dtmf,
 		return;
 
 error:
-	if (req)
-		g_free(req);
+	g_free(req);
 
 	CALLBACK_WITH_FAILURE(cb, data);
 }
@@ -1170,7 +1162,7 @@ static struct ofono_voicecall_driver driver = {
 	.remove			= hfp_voicecall_remove,
 	.dial			= hfp_dial,
 	.answer			= hfp_answer,
-	.hangup			= hfp_hangup,
+	.hangup_active		= hfp_hangup,
 	.hold_all_active	= hfp_hold_all_active,
 	.release_all_held	= hfp_release_all_held,
 	.set_udub		= hfp_set_udub,
